import pytest
from django.contrib.auth import get_user_model
from django.urls import reverse

from ....acl.models import Role
from ....legal.models import Agreement
from ....legal.utils import save_user_agreement_acceptance
from ....test import assert_contains
from ...models import Rank
from ...utils import hash_email

User = get_user_model()


def test_link_is_registered_in_admin_nav(admin_client):
    response = admin_client.get(reverse("misago:admin:index"))
    assert_contains(response, reverse("misago:admin:users:index"))


def test_list_renders_with_item(admin_client, users_admin_link, superuser):
    response = admin_client.get(users_admin_link)
    assert_contains(response, superuser.username)


def test_new_user_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:users:new"))
    assert response.status_code == 200


def test_new_user_can_be_created(admin_client):
    default_rank = Rank.objects.get_default()
    authenticated_role = Role.objects.get(special_role="authenticated")

    admin_client.post(
        reverse("misago:admin:users:new"),
        data={
            "username": "User",
            "rank": str(default_rank.pk),
            "roles": str(authenticated_role.pk),
            "email": "user@example.com",
            "new_password": "pass123",
            "staff_level": "0",
        },
    )

    user = User.objects.get_by_email("user@example.com")
    assert user.username == "User"
    assert user.rank == default_rank
    assert authenticated_role in user.roles.all()
    assert user.check_password("pass123")
    assert not user.is_staff
    assert not user.is_superuser


def test_new_user_can_be_created_with_whitespace_around_password(admin_client):
    default_rank = Rank.objects.get_default()
    authenticated_role = Role.objects.get(special_role="authenticated")

    admin_client.post(
        reverse("misago:admin:users:new"),
        data={
            "username": "User",
            "rank": str(default_rank.pk),
            "roles": str(authenticated_role.pk),
            "email": "user@example.com",
            "new_password": "  pass123  ",
            "staff_level": "0",
        },
    )

    user = User.objects.get_by_email("user@example.com")
    assert user.check_password("  pass123  ")


def test_new_user_creation_fails_because_user_was_not_given_authenticated_role(
    admin_client,
):
    default_rank = Rank.objects.get_default()
    guest_role = Role.objects.get(special_role="anonymous")

    admin_client.post(
        reverse("misago:admin:users:new"),
        data={
            "username": "User",
            "rank": str(default_rank.pk),
            "roles": str(guest_role.pk),
            "email": "user@example.com",
            "new_password": "pass123",
            "staff_level": "0",
        },
    )

    with pytest.raises(User.DoesNotExist):
        User.objects.get_by_email("user@example.com")


def test_edit_user_form_renders(admin_client, user):
    response = admin_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk})
    )
    assert response.status_code == 200


def test_edit_user_form_renders_for_staff_user(staff_client, user):
    response = staff_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk})
    )
    assert response.status_code == 200


def test_edit_staff_form_renders_for_staff_user(staff_client, other_staffuser):
    response = staff_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": other_staffuser.pk})
    )
    assert response.status_code == 200


def test_edit_superuser_form_renders_for_staff_user(staff_client, superuser):
    response = staff_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk})
    )
    assert response.status_code == 200


def get_default_edit_form_data(user):
    default_rank = Rank.objects.get_default()
    authenticated_role = Role.objects.get(special_role="authenticated")
    data = {
        "username": user.username,
        "rank": str(user.rank_id),
        "roles": str(user.roles.all()[0].id),
        "email": user.email,
        "new_password": "",
        "signature": user.signature or "",
        "is_signature_locked": str(user.is_signature_locked),
        "is_hiding_presence": str(user.is_hiding_presence),
        "limits_private_thread_invites_to": str(user.limits_private_thread_invites_to),
        "signature_lock_staff_message": str(user.signature_lock_staff_message or ""),
        "signature_lock_user_message": str(user.signature_lock_user_message or ""),
        "subscribe_to_started_threads": str(user.subscribe_to_started_threads),
        "subscribe_to_replied_threads": str(user.subscribe_to_replied_threads),
        "watch_started_threads": str(user.watch_started_threads),
        "watch_replied_threads": str(user.watch_replied_threads),
        "watch_new_private_threads_by_followed": str(
            user.watch_new_private_threads_by_followed
        ),
        "watch_new_private_threads_by_other_users": str(
            user.watch_new_private_threads_by_other_users
        ),
        "notify_new_private_threads_by_followed": str(
            user.notify_new_private_threads_by_followed
        ),
        "notify_new_private_threads_by_other_users": str(
            user.notify_new_private_threads_by_other_users
        ),
        "is_active": "1",
    }

    if user.is_staff:
        data["is_staff"] = "1"
    if user.is_superuser:
        data["is_superuser"] = "1"

    return data


def test_edit_form_changes_user_username(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["username"] = "NewUsername"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.username == "NewUsername"
    assert user.slug == "newusername"


def test_editing_user_username_creates_entry_in_username_history(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["username"] = "NewUsername"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    assert user.namechanges.exists()


def test_not_editing_user_username_doesnt_create_entry_in_username_history(
    admin_client, user
):
    form_data = get_default_edit_form_data(user)

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    assert not user.namechanges.exists()


def test_edit_form_changes_user_email(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["email"] = "edited@example.com"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.email == "edited@example.com"
    assert user.email_hash == hash_email("edited@example.com")


def test_edit_form_doesnt_remove_current_user_password_if_new_password_is_omitted(
    admin_client, user, user_password
):
    form_data = get_default_edit_form_data(user)

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.check_password(user_password)


def test_edit_form_displays_message_for_user_with_unusable_password(
    admin_client, user, user_password
):
    user.set_password(None)
    user.save()

    response = admin_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk})
    )

    assert_contains(response, "alert-has-unusable-password")


def test_edit_form_doesnt_set_password_for_user_with_unusable_password_if_none_is_given(
    admin_client, user, user_password
):
    user.set_password(None)
    user.save()

    form_data = get_default_edit_form_data(user)

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.has_usable_password()


def test_edit_form_sets_password_for_user_with_unusable_password(
    admin_client, user, user_password
):
    user.set_password(None)
    user.save()

    form_data = get_default_edit_form_data(user)
    form_data["new_password"] = user_password

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.check_password(user_password)


def test_edit_form_changes_user_password(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["new_password"] = "newpassword123"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.check_password("newpassword123")


def test_edit_form_preserves_whitespace_in_new_user_password(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["new_password"] = "  newpassword123  "

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.check_password("  newpassword123  ")


def test_admin_editing_their_own_password_is_not_logged_out(admin_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data["new_password"] = "newpassword123"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    user = admin_client.get("/api/auth/")
    assert user.json()["id"] == superuser.id


def test_staff_user_cannot_degrade_superuser_to_staff_user(staff_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data["is_staff"] = "1"
    form_data.pop("is_superuser")

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_staff
    assert superuser.is_superuser


def test_staff_user_cannot_degrade_superuser_to_regular_user(staff_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data.pop("is_staff")
    form_data.pop("is_superuser")

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_staff
    assert superuser.is_superuser


def test_staff_user_cannot_promote_other_staff_user_to_superuser(
    staff_client, other_staffuser
):
    form_data = get_default_edit_form_data(other_staffuser)
    form_data["is_staff"] = "1"
    form_data["is_superuser"] = "1"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_staffuser.pk}),
        data=form_data,
    )

    other_staffuser.refresh_from_db()
    assert other_staffuser.is_staff
    assert not other_staffuser.is_superuser


def test_staff_user_cannot_promote_regular_user_to_staff(staff_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_staff"] = "1"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.is_staff


def test_staff_user_cannot_promote_regular_user_to_superuser(staff_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_superuser"] = "1"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.is_superuser


def test_staff_user_cannot_promote_themselves_to_superuser(staff_client, staffuser):
    form_data = get_default_edit_form_data(staffuser)
    form_data["is_superuser"] = "1"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert not staffuser.is_superuser


def test_staff_user_cannot_degrade_themselves_to_regular_user(staff_client, staffuser):
    form_data = get_default_edit_form_data(staffuser)
    form_data.pop("is_staff")

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert staffuser.is_staff


def test_superuser_cannot_degrade_themselves_to_staff_user(admin_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data.pop("is_superuser")

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_superuser


def test_superuser_cannot_degrade_themselves_to_regular_user(admin_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data.pop("is_staff")
    form_data.pop("is_superuser")

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_staff
    assert superuser.is_superuser


def test_superuser_can_degrade_other_superuser_to_staff_user(
    admin_client, other_superuser
):
    form_data = get_default_edit_form_data(other_superuser)
    form_data.pop("is_superuser")

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_superuser.pk}),
        data=form_data,
    )

    other_superuser.refresh_from_db()
    assert other_superuser.is_staff
    assert not other_superuser.is_superuser


def test_superuser_can_degrade_other_superuser_to_regular_user(
    admin_client, other_superuser
):
    form_data = get_default_edit_form_data(other_superuser)
    form_data.pop("is_staff")
    form_data.pop("is_superuser")

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_superuser.pk}),
        data=form_data,
    )

    other_superuser.refresh_from_db()
    assert not other_superuser.is_staff
    assert not other_superuser.is_superuser


def test_superuser_can_promote_to_staff_user_to_superuser(admin_client, staffuser):
    form_data = get_default_edit_form_data(staffuser)
    form_data["is_superuser"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert staffuser.is_staff
    assert staffuser.is_superuser


def test_superuser_can_promote_to_regular_user_to_staff_user(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_staff"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.is_staff
    assert not user.is_superuser


def test_superuser_can_promote_to_regular_user_to_superuser(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_staff"] = "1"
    form_data["is_superuser"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.is_staff
    assert user.is_superuser


def test_superuser_can_disable_other_superuser_account(admin_client, other_superuser):
    form_data = get_default_edit_form_data(other_superuser)
    form_data["is_active"] = "0"
    form_data["is_active_staff_message"] = "Test message"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_superuser.pk}),
        data=form_data,
    )

    other_superuser.refresh_from_db()
    assert not other_superuser.is_active
    assert other_superuser.is_active_staff_message == "Test message"


def test_superuser_can_reactivate_other_superuser_account(
    admin_client, other_superuser
):
    other_superuser.is_active = False
    other_superuser.save()

    form_data = get_default_edit_form_data(other_superuser)
    form_data["is_active"] = "1"

    r = admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_superuser.pk}),
        data=form_data,
    )

    other_superuser.refresh_from_db()
    assert other_superuser.is_active


def test_superuser_can_disable_staff_user_account(admin_client, staffuser):
    form_data = get_default_edit_form_data(staffuser)
    form_data["is_active"] = "0"
    form_data["is_active_staff_message"] = "Test message"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert not staffuser.is_active
    assert staffuser.is_active_staff_message == "Test message"


def test_superuser_can_reactivate_staff_user_account(admin_client, staffuser):
    staffuser.is_active = False
    staffuser.save()

    form_data = get_default_edit_form_data(staffuser)
    form_data["is_active"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert staffuser.is_active


def test_superuser_can_disable_regular_user_account(admin_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_active"] = "0"
    form_data["is_active_staff_message"] = "Test message"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.is_active
    assert user.is_active_staff_message == "Test message"


def test_superuser_can_reactivate_regular_user_account(admin_client, user):
    user.is_active = False
    user.save()

    form_data = get_default_edit_form_data(user)
    form_data["is_active"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.is_active


def test_staff_user_can_disable_regular_user_account(staff_client, user):
    form_data = get_default_edit_form_data(user)
    form_data["is_active"] = "0"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.is_active


def test_staff_user_can_reactivate_regular_user_account(staff_client, user):
    user.is_active = False
    user.save()

    form_data = get_default_edit_form_data(user)
    form_data["is_active"] = "1"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert user.is_active


def test_superuser_cant_disable_their_own_account(admin_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data["is_active"] = "0"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_active


def test_staff_user_cant_disable_their_own_account(staff_client, staffuser):
    form_data = get_default_edit_form_data(staffuser)
    form_data["is_active"] = "0"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": staffuser.pk}), data=form_data
    )

    staffuser.refresh_from_db()
    assert staffuser.is_active


def test_staff_user_cant_disable_superuser_account(staff_client, superuser):
    form_data = get_default_edit_form_data(superuser)
    form_data["is_active"] = "0"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": superuser.pk}), data=form_data
    )

    superuser.refresh_from_db()
    assert superuser.is_active


def test_staff_user_cant_disable_other_staff_user_account(
    staff_client, other_staffuser
):
    form_data = get_default_edit_form_data(other_staffuser)
    form_data["is_active"] = "0"

    staff_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": other_staffuser.pk}),
        data=form_data,
    )

    other_staffuser.refresh_from_db()
    assert other_staffuser.is_active


def test_user_deleting_their_account_cant_be_reactivated(admin_client, user):
    user.mark_for_delete()

    form_data = get_default_edit_form_data(user)
    form_data["is_active"] = "1"

    admin_client.post(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk}), data=form_data
    )

    user.refresh_from_db()
    assert not user.is_active


def test_user_agreements_are_displayed_on_edit_form(admin_client, user):
    agreement = Agreement.objects.create(
        type=Agreement.TYPE_TOS,
        title="Test agreement!",
        text="Lorem ipsum!",
        is_active=True,
    )

    save_user_agreement_acceptance(user, agreement, commit=True)

    response = admin_client.get(
        reverse("misago:admin:users:edit", kwargs={"pk": user.pk})
    )
    assert_contains(response, agreement.title)



from django.test import TestCase, override_settings
from django.urls import reverse


class CoreViewsTests(TestCase):
    def test_js_catalog_view_returns_200(self):
        """js catalog view has no show-stoppers"""
        response = self.client.get("/django-i18n.js")
        self.assertEqual(response.status_code, 200)

    def test_robots_txt_returns_200(self):
        """robots.txt has no showstoppers"""
        response = self.client.get("/robots.txt")
        self.assertEqual(response["Content-type"], "text/plain")
        self.assertContains(response, "/api/")


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class RedirectViewTests(TestCase):
    def test_redirect_view(self):
        """redirect view always redirects to home page"""
        response = self.client.get(reverse("test-redirect"))

        self.assertEqual(response.status_code, 302)
        self.assertTrue(response["location"].endswith(reverse("misago:index")))



from typing import List

from django import template
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe

from ..outlets import template_outlets

register = template.Library()

PLUGINOUTLET_SYNTAX_ERROR = 'pluginoutlet tag syntax is "pluginoutlet OUTLET_NAME"'


@register.tag()
def pluginoutlet(_, token):
    split_contents = token.split_contents()
    if len(split_contents) != 2:
        raise template.TemplateSyntaxError(PLUGINOUTLET_SYNTAX_ERROR)

    outlet_name = split_contents[1].strip()
    if not outlet_name:
        raise template.TemplateSyntaxError(PLUGINOUTLET_SYNTAX_ERROR)

    return PluginOutletNode(outlet_name)


class PluginOutletNode(template.Node):
    outlet_name: str

    def __init__(self, outlet_name: str):
        self.outlet_name = outlet_name

    def render(self, context):
        outlet = template_outlets.get(self.outlet_name)
        if not outlet:
            return ""

        request = context.get("request")

        outlet_html = ""
        for plugin_html in outlet(request, context):
            if plugin_html is not None:
                outlet_html += conditional_escape(plugin_html)
        return mark_safe(outlet_html)


HASPLUGINS_SYNTAX_ERROR = 'hasplugins tag syntax is "hasplugins OUTLET_NAME"'


@register.tag()
def hasplugins(parser, token):
    split_contents = token.split_contents()
    if len(split_contents) != 2:
        raise template.TemplateSyntaxError(HASPLUGINS_SYNTAX_ERROR)

    outlet_name = split_contents[1].strip()
    if not outlet_name:
        raise template.TemplateSyntaxError(HASPLUGINS_SYNTAX_ERROR)

    nodelists: List[template.NodeList] = [parser.parse(("else", "endhasplugins"))]
    token = parser.next_token()

    if token.contents == "else":
        nodelists.append(parser.parse(("endhasplugins",)))
        token = parser.next_token()

    if token.contents != "endhasplugins":
        raise template.TemplateSyntaxError(
            'Malformed template tag at line {}: "{}"'.format(
                token.lineno, token.contents
            )
        )

    return HasPluginsNode(outlet_name, nodelists)


class HasPluginsNode(template.Node):
    outlet_name: str
    nodelist: template.NodeList
    nodelist_else: template.NodeList | None

    def __init__(self, outlet_name: str, nodelists: List[template.NodeList]):
        self.outlet_name = outlet_name
        self.nodelist = nodelists[0]

        try:
            self.nodelist_else = nodelists[1]
        except IndexError:
            self.nodelist_else = None

    def render(self, context):
        outlet = template_outlets.get(self.outlet_name)
        if not outlet:
            if self.nodelist_else is not None:
                return self.nodelist_else.render(context)
            return ""

        return self.nodelist.render(context)



import re

from django.conf import settings
from django.db import IntegrityError, models
from django.utils import timezone
from django.utils.translation import pgettext_lazy

from .. import BANS_CACHE
from ...cache.versions import invalidate_cache


class BansManager(models.Manager):
    def get_ip_ban(self, ip, registration_only=False):
        return self.get_ban(ip=ip, registration_only=registration_only)

    def get_username_ban(self, username, registration_only=False):
        return self.get_ban(username=username, registration_only=registration_only)

    def get_email_ban(self, email, registration_only=False):
        return self.get_ban(email=email, registration_only=registration_only)

    def invalidate_cache(self):
        invalidate_cache(BANS_CACHE)

    def get_ban(self, username=None, email=None, ip=None, registration_only=False):
        checks = []

        if username:
            username = username.lower()
            checks.append(self.model.USERNAME)
        if email:
            email = email.lower()
            checks.append(self.model.EMAIL)
        if ip:
            checks.append(self.model.IP)

        queryset = self.filter(is_checked=True)
        if not registration_only:
            queryset = self.filter(registration_only=False)

        if len(checks) == 1:
            queryset = queryset.filter(check_type=checks[0])
        elif checks:
            queryset = queryset.filter(check_type__in=checks)

        for ban in queryset.order_by("-id").iterator():
            if ban.is_expired:
                continue
            elif (
                ban.check_type == self.model.USERNAME
                and username
                and ban.check_value(username)
            ):
                return ban
            elif (
                ban.check_type == self.model.EMAIL and email and ban.check_value(email)
            ):
                return ban
            elif ban.check_type == self.model.IP and ip and ban.check_value(ip):
                return ban

        raise Ban.DoesNotExist("specified values are not banned")


class Ban(models.Model):
    USERNAME = 0
    EMAIL = 1
    IP = 2

    CHOICES = [
        (USERNAME, pgettext_lazy("ban type choice", "Username")),
        (EMAIL, pgettext_lazy("ban type choice", "E-mail address")),
        (IP, pgettext_lazy("ban type choice", "IP address")),
    ]

    check_type = models.PositiveIntegerField(
        default=USERNAME, choices=CHOICES, db_index=True
    )
    registration_only = models.BooleanField(default=False, db_index=True)
    banned_value = models.CharField(max_length=255, db_index=True)
    user_message = models.TextField(null=True, blank=True)
    staff_message = models.TextField(null=True, blank=True)
    expires_on = models.DateTimeField(null=True, blank=True, db_index=True)
    is_checked = models.BooleanField(default=True, db_index=True)

    objects = BansManager()

    def save(self, *args, **kwargs):
        self.banned_value = self.banned_value.lower()
        self.is_checked = not self.is_expired

        return super().save(*args, **kwargs)

    def get_serialized_message(self):
        from ..serializers import BanMessageSerializer

        return BanMessageSerializer(self).data

    @property
    def name(self):
        return self.banned_value

    @property
    def is_expired(self):
        if self.expires_on:
            return self.expires_on < timezone.now()
        return False

    def check_value(self, value):
        if "*" in self.banned_value:
            regex = re.escape(self.banned_value).replace(r"\*", r"(.*?)")
            return re.search("^%s$" % regex, value, re.IGNORECASE) is not None
        return self.banned_value.lower() == value.lower()

    def lift(self):
        self.expires_on = timezone.now()


class BanCache(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        primary_key=True,
        related_name="ban_cache",
        on_delete=models.CASCADE,
    )
    ban = models.ForeignKey(Ban, null=True, blank=True, on_delete=models.SET_NULL)
    cache_version = models.CharField(max_length=8)
    user_message = models.TextField(null=True, blank=True)
    staff_message = models.TextField(null=True, blank=True)
    expires_on = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        try:
            super().save(*args, **kwargs)
        except IntegrityError:
            pass  # first come is first serve with ban cache

    def get_serialized_message(self):
        from ..serializers import BanMessageSerializer

        temp_ban = Ban(
            id=1,
            check_type=Ban.USERNAME,
            user_message=self.user_message,
            staff_message=self.staff_message,
            expires_on=self.expires_on,
        )
        return BanMessageSerializer(temp_ban).data

    @property
    def is_banned(self):
        return bool(self.ban)

    def is_valid(self, cache_versions):
        is_versioned = self.cache_version == cache_versions[BANS_CACHE]
        is_expired = self.expires_on and self.expires_on < timezone.now()

        return is_versioned and not is_expired



from django.contrib.auth import get_user_model
from django.test import TestCase

from ..profilefields import ProfileFields

User = get_user_model()


class ProfileFieldsLoadTests(TestCase):
    def test_no_groups(self):
        """profile fields util handles empty list"""
        profilefields = ProfileFields([])
        profilefields.load()

        self.assertFalse(profilefields.fields_dict)

    def test_empty_group(self):
        """profile fields util handles empty group"""
        profilefields = ProfileFields([{"name": "Test", "fields": []}])

        profilefields.load()

        self.assertFalse(profilefields.fields_dict)

    def test_field_defines_fieldname(self):
        """fields need to define fieldname"""
        profilefields = ProfileFields(
            [
                {
                    "name": "Test",
                    "fields": [
                        "misago.users.tests.testfiles.profilefields.NofieldnameField"
                    ],
                }
            ]
        )

        with self.assertRaises(ValueError):
            profilefields.load()

        try:
            profilefields.load()
        except ValueError as e:
            error = str(e)

            self.assertIn(
                "misago.users.tests.testfiles.profilefields.NofieldnameField", error
            )
            self.assertIn("profile field has to specify fieldname attribute", error)

    def test_detect_repeated_imports(self):
        """fields can't be specified multiple times"""
        profilefields = ProfileFields(
            [
                {
                    "name": "Test",
                    "fields": ["misago.users.profilefields.default.TwitterHandleField"],
                },
                {
                    "name": "Other test",
                    "fields": ["misago.users.profilefields.default.TwitterHandleField"],
                },
            ]
        )

        with self.assertRaises(ValueError):
            profilefields.load()

        try:
            profilefields.load()
        except ValueError as e:
            error = str(e)

            self.assertIn(
                "misago.users.profilefields.default.TwitterHandleField", error
            )
            self.assertIn("profile field has been specified twice", error)

    def test_detect_repeated_fieldnames(self):
        """fields can't reuse other field's fieldnames"""
        profilefields = ProfileFields(
            [
                {
                    "name": "Test",
                    "fields": [
                        "misago.users.tests.testfiles.profilefields.FieldnameField"
                    ],
                },
                {
                    "name": "Other test",
                    "fields": [
                        # pylint: disable=line-too-long
                        "misago.users.tests.testfiles.profilefields.RepeatedFieldnameField"
                    ],
                },
            ]
        )

        with self.assertRaises(ValueError):
            profilefields.load()

        try:
            profilefields.load()
        except ValueError as e:
            error = str(e)

            self.assertIn(
                "misago.users.tests.testfiles.profilefields.FieldnameField", error
            )
            self.assertIn(
                "misago.users.tests.testfiles.profilefields.RepeatedFieldnameField",
                error,
            )
            self.assertIn(
                'field defines fieldname "hello" that is already in use by the', error
            )

    def test_field_correct_field(self):
        """util loads correct field"""
        field_path = "misago.users.profilefields.default.RealNameField"

        profilefields = ProfileFields([{"name": "Test", "fields": [field_path]}])

        profilefields.load()

        self.assertIn(field_path, profilefields.fields_dict)



from django.urls import reverse

from ...notifications.models import WatchedThread
from ...notifications.threads import ThreadNotifications
from ..models import ThreadParticipant


def test_replied_thread_is_watched_by_user_with_option_enabled(
    notify_on_new_thread_reply_mock, user, user_client, thread
):
    user.watch_replied_threads = ThreadNotifications.SITE_ONLY
    user.save()

    response = user_client.post(
        reverse("misago:api:thread-post-list", kwargs={"thread_pk": thread.pk}),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    WatchedThread.objects.get(
        user=user,
        category_id=thread.category_id,
        thread=thread,
        send_emails=False,
    )


def test_replied_thread_is_not_watched_by_user_with_option_disabled(
    notify_on_new_thread_reply_mock, user, user_client, thread
):
    user.watch_replied_threads = ThreadNotifications.NONE
    user.save()

    response = user_client.post(
        reverse("misago:api:thread-post-list", kwargs={"thread_pk": thread.pk}),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200
    assert not WatchedThread.objects.exists()


def test_replied_private_thread_is_watched_by_user_with_option_enabled(
    notify_on_new_thread_reply_mock, user, user_client, private_thread
):
    user.watch_replied_threads = ThreadNotifications.SITE_ONLY
    user.save()

    ThreadParticipant.objects.create(thread=private_thread, user=user)

    response = user_client.post(
        reverse(
            "misago:api:private-thread-post-list",
            kwargs={"thread_pk": private_thread.pk},
        ),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200

    WatchedThread.objects.get(
        user=user,
        category_id=private_thread.category_id,
        thread=private_thread,
        send_emails=False,
    )


def test_replied_private_thread_is_not_watched_by_user_with_option_disabled(
    notify_on_new_thread_reply_mock, user, user_client, private_thread
):
    user.watch_replied_threads = ThreadNotifications.NONE
    user.save()

    ThreadParticipant.objects.create(thread=private_thread, user=user)

    response = user_client.post(
        reverse(
            "misago:api:private-thread-post-list",
            kwargs={"thread_pk": private_thread.pk},
        ),
        data={
            "post": "Lorem ipsum dolor met",
        },
    )

    assert response.status_code == 200
    assert not WatchedThread.objects.exists()



# Generated by Django 3.2.15 on 2023-05-13 20:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("misago_notifications", "0002_migrate_threads_subscriptions"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="notification",
            index=models.Index(
                fields=["-id", "user"], name="misago_noti_id_5a01fa_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="notification",
            index=models.Index(
                fields=["-id", "user", "is_read"], name="misago_noti_id_7b4a20_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="notification",
            index=models.Index(
                condition=models.Q(("is_read", False)),
                fields=["user", "is_read"],
                name="misago_noti_user_unread",
            ),
        ),
        migrations.AddIndex(
            model_name="notification",
            index=models.Index(
                fields=["user", "post", "is_read"],
                name="misago_noti_user_id_9b98fb_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="watchedthread",
            index=models.Index(
                fields=["user", "-thread"], name="misago_noti_user_id_9cde34_idx"
            ),
        ),
    ]



import re

from django import forms
from django.utils.translation import pgettext_lazy

from .base import SettingsForm

GOOGLE_SITE_VERIFICATION = re.compile(
    "^google-site-verification: google([0-9a-z]+)\.html$"
)


class AnalyticsSettingsForm(SettingsForm):
    settings = ["google_tracking_id", "google_site_verification"]

    google_tracking_id = forms.CharField(
        label=pgettext_lazy("admin analytics settings form", "Tracking ID"),
        help_text=pgettext_lazy(
            "admin analytics settings form",
            "Setting the Tracking ID will result in gtag.js file being included in your site's HTML markup, enabling Google Analytics integration.",
        ),
        required=False,
    )
    google_site_verification = forms.CharField(
        label=pgettext_lazy("admin analytics settings form", "Site verification code"),
        help_text=pgettext_lazy(
            "admin analytics settings form",
            "This code was extracted from the uploaded site verification file. To change it, upload new verification file.",
        ),
        required=False,
        disabled=True,
    )
    google_site_verification_file = forms.FileField(
        label=pgettext_lazy(
            "admin analytics settings form", "Upload site verification file"
        ),
        help_text=pgettext_lazy(
            "admin analytics settings form",
            'Site verification file can be downloaded from Search Console\'s "Ownership verification" page.',
        ),
        required=False,
    )

    def clean_google_site_verification_file(self):
        upload = self.cleaned_data.get("google_site_verification_file")
        if not upload:
            return None

        if upload.content_type != "text/html":
            raise forms.ValidationError(
                pgettext_lazy(
                    "admin analytics settings form", "Uploaded file type is not HTML."
                )
            )

        file_content = upload.read().decode("utf-8")
        content_match = GOOGLE_SITE_VERIFICATION.match(file_content)
        if not content_match:
            raise forms.ValidationError(
                pgettext_lazy(
                    "admin analytics settings form",
                    "Uploaded file doesn't contain a verification code.",
                )
            )

        return content_match.group(1)

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("google_site_verification_file"):
            new_verification = cleaned_data.pop("google_site_verification_file")
            cleaned_data["google_site_verification"] = new_verification
        return cleaned_data



from typing import List

from django.http import HttpRequest, HttpResponse, JsonResponse
from rest_framework.decorators import api_view

from ...conf import settings
from ...notifications.models import Notification
from ...notifications.permissions import allow_use_notifications
from ..pagination import paginate_queryset
from .serializers import NotificationSerializer


@api_view(["GET"])
def notifications(request: HttpRequest) -> JsonResponse:
    allow_use_notifications(request.user)

    queryset = (
        Notification.objects.filter(user=request.user)
        .select_related("actor")
        .order_by("-id")
    )

    filter_by = request.GET.get("filter")
    if filter_by == "unread":
        queryset = queryset.filter(is_read=False)
    if filter_by == "read":
        queryset = queryset.filter(is_read=True)

    page = paginate_queryset(
        request,
        queryset,
        "-id",
        settings.MISAGO_NOTIFICATIONS_PAGE_LIMIT,
    )

    # Update user unread notifications counter if its first page of notifications
    # and the counter is obviously invalid
    if (
        "after" not in request.GET
        and "before" not in request.GET
        and filter_by != "read"
    ):
        # Unread notifications counter is non-zero but no notifications exist
        if request.user.unread_notifications and not page.items:
            request.user.unread_notifications = 0
            request.user.save(update_fields=["unread_notifications"])
        # Only page of results, sync unread notifications counter with
        # real count if those don't match
        elif not page.has_next:
            real_unread_notifications = len(
                [
                    notification
                    for notification in page.items
                    if not notification.is_read
                ]
            )
            if real_unread_notifications != request.user.unread_notifications:
                request.user.unread_notifications = real_unread_notifications
                request.user.save(update_fields=["unread_notifications"])

    # Recount unread unread notifications if counter shows 0 but unread
    # notifications exist
    elif not request.user.unread_notifications and unread_items_exist(
        filter_by, page.items
    ):
        count_limit = settings.MISAGO_UNREAD_NOTIFICATIONS_LIMIT + 1
        real_unread_notifications = Notification.objects.filter(
            user=request.user, is_read=False
        )[:count_limit].count()
        if real_unread_notifications:
            request.user.unread_notifications = real_unread_notifications
            request.user.save(update_fields=["unread_notifications"])

    return JsonResponse(
        {
            "results": NotificationSerializer(page.items, many=True).data,
            "hasNext": page.has_next,
            "hasPrevious": page.has_previous,
            "firstCursor": page.first_cursor,
            "lastCursor": page.last_cursor,
            "unreadNotifications": (
                request.user.get_unread_notifications_for_display()
            ),
        }
    )


def unread_items_exist(filter_by: str, items: List[Notification]) -> bool:
    if filter_by == "unread":
        return bool(items)

    if filter_by != "read":
        for notification in items:
            if not notification.is_read:
                return True

    return False


@api_view(["POST"])
def notifications_read_all(request: HttpRequest) -> HttpResponse:
    allow_use_notifications(request.user)

    Notification.objects.filter(
        user=request.user,
        is_read=False,
    ).update(
        is_read=True,
    )
    request.user.unread_notifications = 0
    request.user.save(update_fields=["unread_notifications"])
    return HttpResponse(status=204)



from .. import useracl
from ..test import patch_user_acl


def callable_acl_patch(user, user_acl):
    user_acl["patched_for_user_id"] = user.id


@patch_user_acl({"is_patched": True})
def test_decorator_patches_all_users_acls_in_test(cache_versions, user):
    user_acl = useracl.get_user_acl(user, cache_versions)
    assert user_acl["is_patched"]


def test_decorator_removes_patches_after_test(cache_versions, user):
    @patch_user_acl({"is_patched": True})
    def test_function():
        user_acl = useracl.get_user_acl(user, cache_versions)
        assert user_acl["is_patched"]

    test_function()
    user_acl = useracl.get_user_acl(user, cache_versions)
    assert "is_patched" not in user_acl


def test_context_manager_patches_all_users_acls_in_test(cache_versions, user):
    with patch_user_acl({"can_rename_users": "patched"}):
        user_acl = useracl.get_user_acl(user, cache_versions)
        assert user_acl["can_rename_users"] == "patched"


def test_context_manager_removes_patches_after_exit(cache_versions, user):
    with patch_user_acl({"is_patched": True}):
        user_acl = useracl.get_user_acl(user, cache_versions)
        assert user_acl["is_patched"]

    user_acl = useracl.get_user_acl(user, cache_versions)
    assert "is_patched" not in user_acl


@patch_user_acl(callable_acl_patch)
def test_callable_patch_is_called_with_user_and_acl_by_decorator(cache_versions, user):
    user_acl = useracl.get_user_acl(user, cache_versions)
    assert user_acl["patched_for_user_id"] == user.id


def test_callable_patch_is_called_with_user_and_acl_by_context_manager(
    cache_versions, user
):
    with patch_user_acl(callable_acl_patch):
        user_acl = useracl.get_user_acl(user, cache_versions)
        assert user_acl["patched_for_user_id"] == user.id


@patch_user_acl({"acl_patch": 1})
@patch_user_acl({"acl_patch": 2})
def test_multiple_acl_patches_applied_by_decorator_stack(cache_versions, user):
    user_acl = useracl.get_user_acl(user, cache_versions)
    assert user_acl["acl_patch"] == 2


def test_multiple_acl_patches_applied_by_context_manager_stack(cache_versions, user):
    with patch_user_acl({"acl_patch": 1}):
        with patch_user_acl({"acl_patch": 2}):
            user_acl = useracl.get_user_acl(user, cache_versions)
            assert user_acl["acl_patch"] == 2
        user_acl = useracl.get_user_acl(user, cache_versions)
        assert user_acl["acl_patch"] == 1
    user_acl = useracl.get_user_acl(user, cache_versions)
    assert "acl_patch" not in user_acl



from django import template
from django.utils.translation import pgettext

register = template.Library()


@register.simple_tag
def pagetitle(title, **kwargs):
    if "page" in kwargs and kwargs["page"] > 1:
        title += " (%s)" % (
            pgettext("page title pagination", "page: %(page)s")
            % {"page": kwargs["page"]}
        )

    if "parent" in kwargs:
        title += " | %s" % kwargs["parent"]

    return title



import os

from ariadne import QueryType, load_schema_from_path, make_executable_schema

from .analytics import analytics
from .status import status
from .versioncheck import version_check

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
SCHEMA_PATH = os.path.join(FILE_PATH, "schema.graphql")

type_defs = load_schema_from_path(SCHEMA_PATH)
schema = make_executable_schema(type_defs, [analytics, status, version_check])



import json
from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from .. import test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..models import Post
from ..test import patch_category_acl


class ThreadPostPatchApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)
        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def patch(self, api_link, ops):
        return self.client.patch(
            api_link, json.dumps(ops), content_type="application/json"
        )


class PostAddAclApiTests(ThreadPostPatchApiTestCase):
    def test_add_acl_true(self):
        """api adds current event's acl to response"""
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertTrue(response_json["acl"])

    def test_add_acl_false(self):
        """
        if value is false, api won't add acl to the response, but will set empty key
        """
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIsNone(response_json["acl"])


class PostProtectApiTests(ThreadPostPatchApiTestCase):
    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": True})
    def test_protect_post(self):
        """api makes it possible to protect post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertTrue(reponse_json["is_protected"])

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_protected)

    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": True})
    def test_unprotect_post(self):
        """api makes it possible to unprotect protected post"""
        self.post.is_protected = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_protected"])

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_protected)

    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": True})
    def test_protect_best_answer(self):
        """api makes it possible to protect post"""
        self.thread.set_best_answer(self.user, self.post)
        self.thread.save()

        self.assertFalse(self.thread.best_answer_is_protected)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertTrue(reponse_json["is_protected"])

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_protected)

        self.thread.refresh_from_db()
        self.assertTrue(self.thread.best_answer_is_protected)

    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": True})
    def test_unprotect_best_answer(self):
        """api makes it possible to unprotect protected post"""
        self.post.is_protected = True
        self.post.save()

        self.thread.set_best_answer(self.user, self.post)
        self.thread.save()

        self.assertTrue(self.thread.best_answer_is_protected)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_protected"])

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_protected)

        self.thread.refresh_from_db()
        self.assertFalse(self.thread.best_answer_is_protected)

    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": False})
    def test_protect_post_no_permission(self):
        """api validates permission to protect post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't protect posts in this category."
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_protected)

    @patch_category_acl({"can_edit_posts": 2, "can_protect_posts": False})
    def test_unprotect_post_no_permission(self):
        """api validates permission to unprotect post"""
        self.post.is_protected = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't protect posts in this category."
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_protected)

    @patch_category_acl({"can_edit_posts": 0, "can_protect_posts": True})
    def test_protect_post_not_editable(self):
        """api validates if we can edit post we want to protect"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't protect posts you can't edit."
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_protected)

    @patch_category_acl({"can_edit_posts": 0, "can_protect_posts": True})
    def test_unprotect_post_not_editable(self):
        """api validates if we can edit post we want to protect"""
        self.post.is_protected = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-protected", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't protect posts you can't edit."
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_protected)


class PostApproveApiTests(ThreadPostPatchApiTestCase):
    @patch_category_acl({"can_approve_content": True})
    def test_approve_post(self):
        """api makes it possible to approve post"""
        self.post.is_unapproved = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_unapproved"])

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": True})
    def test_unapprove_post(self):
        """unapproving posts is not supported by api"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "Content approval can't be reversed."
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": False})
    def test_approve_post_no_permission(self):
        """api validates approval permission"""
        self.post.is_unapproved = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't approve posts in this category."
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": True, "can_close_threads": False})
    def test_approve_post_closed_thread_no_permission(self):
        """api validates approval permission in closed threads"""
        self.post.is_unapproved = True
        self.post.save()

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't approve posts in it.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": True, "can_close_threads": False})
    def test_approve_post_closed_category_no_permission(self):
        """api validates approval permission in closed categories"""
        self.post.is_unapproved = True
        self.post.save()

        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't approve posts in it.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": True})
    def test_approve_first_post(self):
        """api approve first post fails"""
        self.post.is_unapproved = True
        self.post.save()

        self.thread.set_first_post(self.post)
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "Thread's first post can only be approved together with thread.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_unapproved)

    @patch_category_acl({"can_approve_content": True})
    def test_approve_hidden_post(self):
        """api approve hidden post fails"""
        self.post.is_unapproved = True
        self.post.is_hidden = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-unapproved", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't approve posts the content you can't see.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_unapproved)


class PostHideApiTests(ThreadPostPatchApiTestCase):
    @patch_category_acl({"can_hide_posts": 1})
    def test_hide_post(self):
        """api makes it possible to hide post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertTrue(reponse_json["is_hidden"])

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_hide_posts": 1})
    def test_hide_own_post(self):
        """api makes it possible to hide owned post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertTrue(reponse_json["is_hidden"])

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_hide_posts": 0})
    def test_hide_post_no_permission(self):
        """api hide post with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't hide posts in this category."
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_hide_own_posts": 1, "can_protect_posts": False})
    def test_hide_own_protected_post(self):
        """api validates if we are trying to hide protected post"""
        self.post.is_protected = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This post is protected. You can't hide it."
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_hide_own_posts": True})
    def test_hide_other_user_post(self):
        """api validates post ownership when hiding"""
        self.post.poster = None
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't hide other users posts in this category.",
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"post_edit_time": 1, "can_hide_own_posts": True})
    def test_hide_own_post_after_edit_time(self):
        """api validates if we are trying to hide post after edit time"""
        self.post.posted_on = timezone.now() - timedelta(minutes=10)
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't hide posts that are older than 1 minute.",
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_own_posts": True})
    def test_hide_post_in_closed_thread(self):
        """api validates if we are trying to hide post in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't hide posts in it.",
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_own_posts": True})
    def test_hide_post_in_closed_category(self):
        """api validates if we are trying to hide post in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't hide posts in it.",
        )

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_hide_posts": 1})
    def test_hide_first_post(self):
        """api hide first post fails"""
        self.thread.set_first_post(self.post)
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "Thread's first post can only be hidden using the hide thread option.",
        )

    @patch_category_acl({"can_hide_posts": 1})
    def test_hide_best_answer(self):
        """api hide first post fails"""
        self.thread.set_best_answer(self.user, self.post)
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.post.id,
                "detail": [
                    "You can't hide this post because its marked as best answer."
                ],
            },
        )


class PostUnhideApiTests(ThreadPostPatchApiTestCase):
    @patch_category_acl({"can_hide_posts": 1})
    def test_show_post(self):
        """api makes it possible to unhide post"""
        self.post.is_hidden = True
        self.post.save()

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_hidden"])

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_hide_own_posts": 1})
    def test_show_own_post(self):
        """api makes it possible to unhide owned post"""
        self.post.is_hidden = True
        self.post.save()

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertFalse(reponse_json["is_hidden"])

        self.post.refresh_from_db()
        self.assertFalse(self.post.is_hidden)

    @patch_category_acl({"can_hide_posts": 0})
    def test_show_post_no_permission(self):
        """api unhide post with no permission fails"""
        self.post.is_hidden = True
        self.post.save()

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't reveal posts in this category."
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_protect_posts": 0, "can_hide_own_posts": 1})
    def test_show_own_protected_post(self):
        """api validates if we are trying to reveal protected post"""
        self.post.is_hidden = True
        self.post.save()

        self.post.is_protected = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "This post is protected. You can't reveal it."
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_hide_own_posts": 1})
    def test_show_other_user_post(self):
        """api validates post ownership when revealing"""
        self.post.is_hidden = True
        self.post.poster = None
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't reveal other users posts in this category.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"post_edit_time": 1, "can_hide_own_posts": 1})
    def test_show_own_post_after_edit_time(self):
        """api validates if we are trying to reveal post after edit time"""
        self.post.is_hidden = True
        self.post.posted_on = timezone.now() - timedelta(minutes=10)
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "You can't reveal posts that are older than 1 minute.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_own_posts": 1})
    def test_show_post_in_closed_thread(self):
        """api validates if we are trying to reveal post in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        self.post.is_hidden = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't reveal posts in it.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_own_posts": 1})
    def test_show_post_in_closed_category(self):
        """api validates if we are trying to reveal post in closed category"""
        self.category.is_closed = True
        self.category.save()

        self.post.is_hidden = True
        self.post.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't reveal posts in it.",
        )

        self.post.refresh_from_db()
        self.assertTrue(self.post.is_hidden)

    @patch_category_acl({"can_hide_posts": 1})
    def test_show_first_post(self):
        """api unhide first post fails"""
        self.thread.set_first_post(self.post)
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "Thread's first post can only be revealed using the reveal thread option.",
        )


class PostLikeApiTests(ThreadPostPatchApiTestCase):
    @patch_category_acl({"can_see_posts_likes": 0})
    def test_like_no_see_permission(self):
        """api validates user's permission to see posts likes"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": True}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.post.id, "detail": ["You can't like posts in this category."]},
        )

    @patch_category_acl({"can_like_posts": False})
    def test_like_no_like_permission(self):
        """api validates user's permission to see posts likes"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": True}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.post.id, "detail": ["You can't like posts in this category."]},
        )

    def test_like_post(self):
        """api adds user like to post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["likes"], 1)
        self.assertEqual(response_json["is_liked"], True)
        self.assertEqual(
            response_json["last_likes"],
            [{"id": self.user.id, "username": self.user.username}],
        )

        post = Post.objects.get(pk=self.post.pk)
        self.assertEqual(post.likes, response_json["likes"])
        self.assertEqual(post.last_likes, response_json["last_likes"])

    def test_like_liked_post(self):
        """api adds user like to post"""
        test.like_post(self.post, username="Myo")
        test.like_post(self.post, username="Mugi")
        test.like_post(self.post, username="Bob")
        test.like_post(self.post, username="Miku")

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["likes"], 5)
        self.assertEqual(response_json["is_liked"], True)
        self.assertEqual(
            response_json["last_likes"],
            [
                {"id": self.user.id, "username": self.user.username},
                {"id": None, "username": "Miku"},
                {"id": None, "username": "Bob"},
                {"id": None, "username": "Mugi"},
            ],
        )

        post = Post.objects.get(pk=self.post.pk)
        self.assertEqual(post.likes, response_json["likes"])
        self.assertEqual(post.last_likes, response_json["last_likes"])

    def test_unlike_post(self):
        """api removes user like from post"""
        test.like_post(self.post, self.user)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["likes"], 0)
        self.assertEqual(response_json["is_liked"], False)
        self.assertEqual(response_json["last_likes"], [])

        post = Post.objects.get(pk=self.post.pk)
        self.assertEqual(post.likes, response_json["likes"])
        self.assertEqual(post.last_likes, response_json["last_likes"])

    def test_like_post_no_change(self):
        """api does no state change if we are linking liked post"""
        test.like_post(self.post, self.user)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["likes"], 1)
        self.assertEqual(response_json["is_liked"], True)
        self.assertEqual(
            response_json["last_likes"],
            [{"id": self.user.id, "username": self.user.username}],
        )

        post = Post.objects.get(pk=self.post.pk)
        self.assertEqual(post.likes, response_json["likes"])
        self.assertEqual(post.last_likes, response_json["last_likes"])

    def test_unlike_post_no_change(self):
        """api does no state change if we are unlinking unliked post"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-liked", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["likes"], 0)
        self.assertEqual(response_json["is_liked"], False)
        self.assertEqual(response_json["last_likes"], [])


class ThreadEventPatchApiTestCase(ThreadPostPatchApiTestCase):
    def setUp(self):
        super().setUp()

        self.event = test.reply_thread(self.thread, poster=self.user, is_event=True)

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.event.pk},
        )

    def refresh_event(self):
        self.event = self.thread.post_set.get(pk=self.event.pk)


class EventAnonPatchApiTests(ThreadEventPatchApiTestCase):
    def test_anonymous_user(self):
        """anonymous users can't change event state"""
        self.logout_user()

        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": True}]
        )
        self.assertEqual(response.status_code, 403)


class EventAddAclApiTests(ThreadEventPatchApiTestCase):
    def test_add_acl_true(self):
        """api adds current event's acl to response"""
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertTrue(response_json["acl"])

    def test_add_acl_false(self):
        """
        if value is false, api won't add acl to the response, but will set empty key
        """
        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIsNone(response_json["acl"])

        response = self.patch(
            self.api_link, [{"op": "add", "path": "acl", "value": True}]
        )
        self.assertEqual(response.status_code, 200)


class EventHideApiTests(ThreadEventPatchApiTestCase):
    @patch_category_acl({"can_hide_events": 1})
    def test_hide_event(self):
        """api makes it possible to hide event"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 200)

        self.event.refresh_from_db()
        self.assertTrue(self.event.is_hidden)

    @patch_category_acl({"can_hide_events": 1})
    def test_show_event(self):
        """api makes it possible to unhide event"""
        self.event.is_hidden = True
        self.event.save()

        self.event.refresh_from_db()
        self.assertTrue(self.event.is_hidden)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 200)

        self.event.refresh_from_db()
        self.assertFalse(self.event.is_hidden)

    @patch_category_acl({"can_hide_events": 0})
    def test_hide_event_no_permission(self):
        """api hide event with no permission fails"""
        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0], "You can't hide events in this category."
        )

        self.event.refresh_from_db()
        self.assertFalse(self.event.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_events": 1})
    def test_hide_event_closed_thread_no_permission(self):
        """api hide event in closed thread with no permission fails"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't hide events in it.",
        )

        self.event.refresh_from_db()
        self.assertFalse(self.event.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_events": 1})
    def test_hide_event_closed_category_no_permission(self):
        """api hide event in closed category with no permission fails"""
        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": True}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't hide events in it.",
        )

        self.event.refresh_from_db()
        self.assertFalse(self.event.is_hidden)

    @patch_category_acl({"can_hide_events": 0})
    def test_show_event_no_permission(self):
        """api unhide event with no permission fails"""
        self.event.is_hidden = True
        self.event.save()

        self.event.refresh_from_db()
        self.assertTrue(self.event.is_hidden)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_close_threads": False, "can_hide_events": 1})
    def test_show_event_closed_thread_no_permission(self):
        """api show event in closed thread with no permission fails"""
        self.event.is_hidden = True
        self.event.save()

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This thread is closed. You can't reveal events in it.",
        )

        self.event.refresh_from_db()
        self.assertTrue(self.event.is_hidden)

    @patch_category_acl({"can_close_threads": False, "can_hide_events": 1})
    def test_show_event_closed_category_no_permission(self):
        """api show event in closed category with no permission fails"""
        self.event.is_hidden = True
        self.event.save()

        self.category.is_closed = True
        self.category.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "is-hidden", "value": False}]
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["detail"][0],
            "This category is closed. You can't reveal events in it.",
        )

        self.event.refresh_from_db()
        self.assertTrue(self.event.is_hidden)



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("misago_threads", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="CategoryRead",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("last_read_on", models.DateTimeField()),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ThreadRead",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("last_read_on", models.DateTimeField()),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_categories.Category",
                    ),
                ),
                (
                    "thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="misago_threads.Thread",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



from unittest.mock import Mock

from ..context_processors import user_acl


def test_context_processor_adds_request_user_acl_to_context():
    test_acl = {"test": True}
    context = user_acl(Mock(user_acl=test_acl))
    assert context == {"user_acl": test_acl}



from django import forms
from django.utils.translation import pgettext_lazy

from ...acl import algebra
from ...acl.models import Role
from ...admin.forms import YesNoSwitch


class PermissionsForm(forms.Form):
    legend = pgettext_lazy("users accounts permission", "Account settings")

    name_changes_allowed = forms.IntegerField(
        label=pgettext_lazy(
            "users accounts permission", "Allowed username changes number"
        ),
        min_value=0,
        initial=1,
    )
    name_changes_expire = forms.IntegerField(
        label=pgettext_lazy(
            "users accounts permission", "Don't count username changes older than"
        ),
        help_text=pgettext_lazy(
            "users accounts permission",
            "Number of days since name change that makes that change no longer count to limit. Enter zero to make all changes count.",
        ),
        min_value=0,
        initial=0,
    )
    can_have_signature = YesNoSwitch(
        label=pgettext_lazy("users accounts permission", "Can have signature")
    )
    allow_signature_links = YesNoSwitch(
        label=pgettext_lazy("users accounts permission", "Can put links in signature")
    )
    allow_signature_images = YesNoSwitch(
        label=pgettext_lazy("users accounts permission", "Can put images in signature")
    )
    allow_signature_blocks = YesNoSwitch(
        label=pgettext_lazy(
            "users accounts permission", "Can use text blocks in signature"
        ),
        help_text=pgettext_lazy(
            "users accounts permission",
            "Controls whether or not users can put quote, code, spoiler blocks and horizontal lines in signatures.",
        ),
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "name_changes_allowed": 0,
        "name_changes_expire": 0,
        "can_have_signature": 0,
        "allow_signature_links": 0,
        "allow_signature_images": 0,
        "allow_signature_blocks": 0,
    }
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        name_changes_allowed=algebra.greater,
        name_changes_expire=algebra.lower_non_zero,
        can_have_signature=algebra.greater,
        allow_signature_links=algebra.greater,
        allow_signature_images=algebra.greater,
        allow_signature_blocks=algebra.greater,
    )



from django.core.management.base import BaseCommand
from django.utils import timezone

from ....core.pgutils import chunk_queryset
from ...datadownloads import expire_user_data_download
from ...models import DataDownload


class Command(BaseCommand):
    help = "Expires old user data downloads."

    def handle(self, *args, **options):
        downloads_expired = 0
        queryset = DataDownload.objects.select_related("user")
        queryset = queryset.filter(
            status=DataDownload.STATUS_READY, expires_on__lte=timezone.now()
        )

        for data_download in chunk_queryset(queryset):
            expire_user_data_download(data_download)
            downloads_expired += 1

        self.stdout.write("Data downloads expired: %s" % downloads_expired)



from io import StringIO

from django.core.management import call_command

from ...threads.models import Post
from ..management.commands import createfakeposts
from ..threads import get_fake_thread


def test_management_command_creates_fake_threads(fake, default_category):
    thread = get_fake_thread(fake, default_category)
    call_command(createfakeposts.Command(), stdout=StringIO())
    assert Post.objects.exclude(pk=thread.first_post.pk).exists()



import requests
from celery import shared_task
from requests.exceptions import RequestException

from ..cache import clear_theme_cache
from ..models import Theme, Css
from .css import get_theme_media_map, rebuild_css


@shared_task
def update_remote_css_size(pk):
    try:
        css = Css.objects.get(pk=pk, url__isnull=False)
    except Css.DoesNotExist:
        pass
    else:
        css.size = get_remote_css_size(css.url)
        css.save(update_fields=["size"])


def get_remote_css_size(url):
    try:
        with requests.get(url, stream=True) as response:
            response.raise_for_status()
            return sum(len(chunk) for chunk in response.iter_content(4096))
    except RequestException:
        return 0


@shared_task
def build_single_theme_css(pk):
    try:
        css = Css.objects.get(pk=pk, source_needs_building=True)
    except Css.DoesNotExist:
        pass
    else:
        media_map = get_theme_media_map(css.theme)
        rebuild_css(media_map, css)
        clear_theme_cache()


@shared_task
def build_theme_css(pk):
    try:
        theme = Theme.objects.get(pk=pk)
    except Theme.DoesNotExist:
        pass
    else:
        media_map = get_theme_media_map(theme)
        for css in theme.css.filter(source_needs_building=True):
            rebuild_css(media_map, css)
        clear_theme_cache()



import json

from ..useracl import get_user_acl, serialize_user_acl


def test_user_acl_is_serializeable(cache_versions, user):
    acl = get_user_acl(user, cache_versions)
    assert serialize_user_acl(acl)


def test_user_acl_is_json_serializeable(cache_versions, user):
    acl = get_user_acl(user, cache_versions)
    serialized_acl = serialize_user_acl(acl)
    assert json.dumps(serialized_acl)



from ...core.apirouter import MisagoApiRouter
from ..api.attachments import AttachmentViewSet
from ..api.threadpoll import ThreadPollViewSet
from ..api.threadposts import PrivateThreadPostsViewSet, ThreadPostsViewSet
from ..api.threads import PrivateThreadViewSet, ThreadViewSet

router = MisagoApiRouter()

router.register(r"attachments", AttachmentViewSet, basename="attachment")

router.register(r"threads", ThreadViewSet, basename="thread")
router.register(
    r"threads/(?P<thread_pk>[^/.]+)/posts", ThreadPostsViewSet, basename="thread-post"
)
router.register(
    r"threads/(?P<thread_pk>[^/.]+)/poll", ThreadPollViewSet, basename="thread-poll"
)

router.register(r"private-threads", PrivateThreadViewSet, basename="private-thread")
router.register(
    r"private-threads/(?P<thread_pk>[^/.]+)/posts",
    PrivateThreadPostsViewSet,
    basename="private-thread-post",
)

urlpatterns = router.urls



from .analytics import AnalyticsSettingsForm
from .base import SettingsForm
from .captcha import CaptchaSettingsForm
from .general import GeneralSettingsForm
from .notifications import NotificationsSettingsForm
from .oauth2 import OAuth2SettingsForm
from .threads import ThreadsSettingsForm
from .users import UsersSettingsForm



from django.test import TestCase

from ...cache.models import CacheVersion
from ..pgutils import chunk_queryset


class ChunkQuerysetTest(TestCase):
    def setUp(self):
        # clear table
        CacheVersion.objects.all().delete()

        items_pks = []
        for i in range(50):
            obj = CacheVersion.objects.create(cache="test%s" % i)
            items_pks.append(obj.pk)
        self.items_pks = list(sorted(items_pks, reverse=True))

    def test_chunk_queryset(self):
        """chunk_queryset utility chunks queryset but returns all items"""
        chunked_pks = []

        with self.assertNumQueries(1):
            queryset = CacheVersion.objects.order_by("cache")
            for obj in chunk_queryset(queryset, chunk_size=5):
                chunked_pks.append(obj.pk)

        self.assertEqual(chunked_pks, self.items_pks)

    def test_chunk_shrinking_queryset(self):
        """chunk_queryset utility chunks queryset in delete action"""
        with self.assertNumQueries(51):
            queryset = CacheVersion.objects.all()
            for obj in chunk_queryset(queryset, chunk_size=5):
                obj.delete()

        self.assertEqual(CacheVersion.objects.count(), 0)



from django.db import models

from .utils import generate_version_string


class CacheVersion(models.Model):
    cache = models.CharField(max_length=128, primary_key=True)
    version = models.CharField(max_length=8, default=generate_version_string)



from django import forms
from django.utils.translation import pgettext_lazy

from ....admin.forms import YesNoSwitch
from .base import SettingsForm


class CaptchaSettingsForm(SettingsForm):
    settings = [
        "captcha_type",
        "recaptcha_site_key",
        "recaptcha_secret_key",
        "qa_question",
        "qa_help_text",
        "qa_answers",
        "enable_stop_forum_spam",
        "stop_forum_spam_confidence",
    ]

    captcha_type = forms.ChoiceField(
        label=pgettext_lazy("admin captcha settings form", "Enable CAPTCHA"),
        choices=[
            ("no", pgettext_lazy("admin captcha type field choice", "No CAPTCHA")),
            ("re", pgettext_lazy("admin captcha type field choice", "reCaptcha")),
            (
                "qa",
                pgettext_lazy("admin captcha type field choice", "Question and answer"),
            ),
        ],
        widget=forms.RadioSelect(),
    )

    recaptcha_site_key = forms.CharField(
        label=pgettext_lazy("admin captcha settings form", "Site key"),
        max_length=100,
        required=False,
    )
    recaptcha_secret_key = forms.CharField(
        label=pgettext_lazy("admin captcha settings form", "Secret key"),
        max_length=100,
        required=False,
    )

    qa_question = forms.CharField(
        label=pgettext_lazy("admin captcha settings form", "Test question"),
        max_length=100,
        required=False,
    )
    qa_help_text = forms.CharField(
        label=pgettext_lazy("admin captcha settings form", "Question help text"),
        max_length=250,
        required=False,
    )
    qa_answers = forms.CharField(
        label=pgettext_lazy("admin captcha settings form", "Valid answers"),
        help_text=pgettext_lazy(
            "admin captcha settings form",
            "Enter each answer in new line. Answers are case-insensitive.",
        ),
        widget=forms.Textarea({"rows": 4}),
        max_length=250,
        required=False,
    )

    enable_stop_forum_spam = YesNoSwitch(
        label=pgettext_lazy(
            "admin captcha settings form",
            "Validate new registrations against SFS database",
        ),
        help_text=pgettext_lazy(
            "admin captcha settings form",
            "Turning this option on will result in Misago validating new user's e-mail and IP address against SFS database.",
        ),
    )
    stop_forum_spam_confidence = forms.IntegerField(
        label=pgettext_lazy(
            "admin captcha settings form", "Minimum SFS confidence required"
        ),
        help_text=pgettext_lazy(
            "admin captcha settings form",
            "SFS compares user e-mail and IP address with database of known spammers and assigns the confidence score in range of 0 to 100 that user is a spammer themselves. If this score is equal or higher than specified, Misago will block user from registering and ban their IP address for 24 hours.",
        ),
        min_value=0,
        max_value=100,
    )

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data.get("captcha_type") == "re":
            if not cleaned_data.get("recaptcha_site_key"):
                self.add_error(
                    "recaptcha_site_key",
                    pgettext_lazy(
                        "admin captcha settings form",
                        "You need to enter site key if selected CAPTCHA type is reCaptcha.",
                    ),
                )

            if not cleaned_data.get("recaptcha_secret_key"):
                self.add_error(
                    "recaptcha_secret_key",
                    pgettext_lazy(
                        "admin captcha settings form",
                        "You need to enter secret key if selected CAPTCHA type is reCaptcha.",
                    ),
                )

        if cleaned_data.get("captcha_type") == "qa":
            if not cleaned_data.get("qa_question"):
                self.add_error(
                    "qa_question",
                    pgettext_lazy(
                        "admin captcha settings form",
                        "You need to set question if selected CAPTCHA type is Q&A.",
                    ),
                )

            if not cleaned_data.get("qa_answers"):
                self.add_error(
                    "qa_answers",
                    pgettext_lazy(
                        "admin captcha settings form",
                        "You need to set question answers if selected CAPTCHA type is Q&A.",
                    ),
                )

        return cleaned_data



from django.test import TestCase
from django.urls import reverse

from ..test import create_test_user


class AuthenticateApiTests(TestCase):
    def setUp(self):
        self.api_link = reverse("misago:api:mention-suggestions")

    def test_no_query(self):
        """api returns empty result set if no query is given"""
        response = self.client.get(self.api_link)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])

    def test_no_results(self):
        """api returns empty result set if no query is given"""
        response = self.client.get(self.api_link + "?q=none")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])

    def test_user_search(self):
        """api searches uses"""
        create_test_user("User", "user@example.com")

        # exact case sensitive match
        response = self.client.get(self.api_link + "?q=User")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [{"avatar": "http://dummyimage.com/100/100", "username": "User"}],
        )

        # case insensitive match
        response = self.client.get(self.api_link + "?q=user")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [{"avatar": "http://dummyimage.com/100/100", "username": "User"}],
        )

        # eager case insensitive match
        response = self.client.get(self.api_link + "?q=u")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [{"avatar": "http://dummyimage.com/100/100", "username": "User"}],
        )

        # invalid match
        response = self.client.get(self.api_link + "?q=other")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])



from django.test import TestCase
from django.utils import translation

from .. import context_processors


class MockRequest:
    path = "/"

    def __init__(self, secure, host):
        self.secure = secure
        self.host = host

    def is_secure(self):
        return self.secure

    def get_host(self):
        return self.host


class MetaMockRequest:
    def __init__(self, meta):
        self.META = meta


class MomentjsLocaleTests(TestCase):
    def test_momentjs_locale(self):
        """momentjs_locale adds MOMENTJS_LOCALE_URL to context"""
        with translation.override("no-no"):
            self.assertEqual(
                context_processors.momentjs_locale(True), {"MOMENTJS_LOCALE_URL": None}
            )

        with translation.override("en-us"):
            self.assertEqual(
                context_processors.momentjs_locale(True), {"MOMENTJS_LOCALE_URL": None}
            )

        with translation.override("de"):
            self.assertEqual(
                context_processors.momentjs_locale(True),
                {"MOMENTJS_LOCALE_URL": "misago/momentjs/de.js"},
            )

        with translation.override("pl-de"):
            self.assertEqual(
                context_processors.momentjs_locale(True),
                {"MOMENTJS_LOCALE_URL": "misago/momentjs/pl.js"},
            )


class FrontendContextTests(TestCase):
    def test_frontend_context(self):
        """frontend_context is available in templates"""
        mock_request = MockRequest(False, "somewhere.com")
        mock_request.include_frontend_context = True
        mock_request.frontend_context = {"someValue": "Something"}

        self.assertEqual(
            context_processors.frontend_context(mock_request),
            {"frontend_context": {"someValue": "Something"}},
        )

        mock_request.include_frontend_context = False
        self.assertEqual(context_processors.frontend_context(mock_request), {})



from unittest.mock import Mock

import pytest

from ...users import signatures


@pytest.fixture
def signature(user):
    user.signature = "Test"
    user.signature_parsed = "Test"
    user.signature_checksum = "Test"
    user.save()


def test_user_signature_and_valid_checksum_is_set(user, signature, user_acl):
    request = Mock(scheme="http", get_host=Mock(return_value="127.0.0.1:800"))
    signatures.set_user_signature(request, user, user_acl, "Changed")

    assert user.signature == "Changed"
    assert user.signature_parsed == "<p>Changed</p>"
    assert user.signature_checksum
    assert signatures.is_user_signature_valid(user)


def test_user_signature_is_cleared(user, signature, user_acl):
    request = Mock(scheme="http", get_host=Mock(return_value="127.0.0.1:800"))
    signatures.set_user_signature(request, user, user_acl, "")

    assert not user.signature
    assert not user.signature_parsed
    assert not user.signature_checksum


def test_signature_validity_check_fails_for_incorrect_signature_checksum(
    user, signature
):
    assert not signatures.is_user_signature_valid(user)
